/**
 * Diashapes -- Assistant to download and install Dia shapes
 * Copyright (C) 2007, 2008 Steffen Macke <sdteffen@sdteffen.de>
 *
 * main.c: The main function
 *
 * Format with indent -bli 0
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libxml/nanohttp.h>
#include <libxml/xmlreader.h>
#include <curl/curl.h>
#include <gsf/gsf.h>

#ifdef G_OS_WIN32
#include <windows.h>
#include <stdio.h>
#endif

#include "diashapes.h"

static int verbose = 0;

static CURL *curl;

static gpointer zip_in_memory = NULL;

static gint zip_size = 0;

static GtkTreeModel *
create_and_fill_model (gchar * sheetpath, GtkWindow * assistant)
{
  GtkListStore *store = NULL;
  GtkTreeIter iter;
  xmlDoc *doc = NULL;
  xmlNode *root_element = NULL;
  xmlNode *sheet_node = NULL;
  GtkMessageDialog *dialog = NULL;

  LIBXML_TEST_VERSION
    /*parse the file and get the DOM */
    doc = xmlReadFile (sheetpath, NULL, 0);

  if (doc == NULL)
  {
    dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (assistant,
							 GTK_DIALOG_DESTROY_WITH_PARENT,
							 GTK_MESSAGE_ERROR,
							 GTK_BUTTONS_CLOSE,
							 "error: could not parse file %s\n",
							 sheetpath));
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (GTK_WIDGET (dialog));

  }

  /*Get the root element node */
  root_element = xmlDocGetRootElement (doc);

  store = gtk_list_store_new (NUM_COLS, G_TYPE_STRING, G_TYPE_STRING,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

  if (NULL != root_element->children)
  {
    for (sheet_node = root_element->children; sheet_node; sheet_node
	 = sheet_node->next)
    {
      if ((XML_ELEMENT_NODE == sheet_node->type)
	  && (0 == g_ascii_strncasecmp (sheet_node->name, "sheet", 5)))
      {
	gtk_list_store_append (store, &iter);
	gtk_list_store_set (store, &iter, COL_NAME,
			    xmlGetProp (sheet_node, "name"),
			    COL_DESCRIPTION, xmlGetProp (sheet_node,
							 "description"),
			    COL_CREATOR, xmlGetProp (sheet_node,
						     "creator"),
			    COL_WEBSITE, xmlGetProp (sheet_node,
						     "website"),
			    COL_DOWNLOAD, xmlGetProp (sheet_node,
						      "download"), -1);
      }
    }
  }

  /*free the document */
  xmlFreeDoc (doc);

  /*
   *Free the global variables that may
   *have been allocated by the parser.
   */
  xmlCleanupParser ();

  return GTK_TREE_MODEL (store);
}

static GtkWidget *
create_view_and_model (gchar * sheetpath, GtkWindow * assistant)
{
  GtkCellRenderer *renderer;
  GtkTreeModel *model;
  GtkWidget *view;
  GtkTreeSelection *selection;

  view = gtk_tree_view_new ();

  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       -1,
					       "Name",
					       renderer,
					       "text", COL_NAME, NULL);
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       -1,
					       "Description",
					       renderer,
					       "text", COL_DESCRIPTION, NULL);
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       -1,
					       "Creator",
					       renderer,
					       "text", COL_CREATOR, NULL);
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       -1,
					       "Website",
					       renderer,
					       "text", COL_WEBSITE, NULL);
  model = create_and_fill_model (sheetpath, assistant);

  gtk_tree_view_set_model (GTK_TREE_VIEW (view), model);

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);

  /* The tree view has acquired its own reference to the
   *  model, so we can drop ours. That way the model will
   *  be freed automatically when the tree view is destroyed */

  g_object_unref (model);

  return view;
}

/**
 * If there is a selection in the GtkTreeView, set the page as complete. 
 * Otherwise stop the user from progressing the next page. 
 */
static void
cursor_changed (GtkTreeView * view, GtkAssistant * assistant)
{
  GtkTreeSelection *selection = gtk_tree_view_get_selection (view);
  gint num = gtk_assistant_get_current_page (assistant);
  GtkWidget *page = gtk_assistant_get_nth_page (assistant, num);
  gtk_assistant_set_page_complete (assistant, page, (0
						     <
						     gtk_tree_selection_count_selected_rows
						     (selection)));
}

/**
 * If the dialog is cancelled, delete it and clean up. 
 */
static void
assistant_cancel (GtkAssistant * assistant, gpointer data)
{
  gtk_widget_destroy (GTK_WIDGET (assistant));
}

/** 
 * Handle page-specific actions 
 */
static void
assistant_prepare (GtkAssistant * assistant, GtkWidget * page, gpointer data)
{
  GtkProgressBar *progress;
  gint page_number = gtk_assistant_get_current_page (assistant);

  /**
   * \TODO: Avoid clicking on "INSTALL" button
   */
  if (3 == page_number)
  {

  }
}

/**
 * Fetch the sheets
 */
static void
fetch_sheets (GtkWidget * progressbar, gpointer page)
{
  GList *selected_rows = NULL;
  guint selected_row_count = 0;
  guint processed_row_count = 0;
  GtkProgressBar *progress = NULL;
  GtkTreeView *view = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreePath *selected_row_path = NULL;
  GtkTreeIter iter;
  GtkTreeModel *model = NULL;
  gchar *sheet_url = NULL;
  gchar *sheet_name = NULL;
  gchar *message = NULL;
  gdouble percent = 0.0;
  GtkAssistant *assistant = NULL;
  GtkLabel *label = NULL;
  GtkLabel *summary_label = NULL;

  g_chdir (get_dia_dir ());
  view = GTK_TREE_VIEW (g_object_get_data (G_OBJECT (page), "view"));
  assistant =
    GTK_ASSISTANT (g_object_get_data (G_OBJECT (page), "assistant"));
  label = GTK_LABEL (g_object_get_data (G_OBJECT (page), "label"));
  model = gtk_tree_view_get_model (view);
  selection = gtk_tree_view_get_selection (view);
  progress = GTK_PROGRESS_BAR (g_object_get_data (G_OBJECT (page), "pbar"));
  selected_rows = gtk_tree_selection_get_selected_rows (selection, NULL);
  selected_row_count = gtk_tree_selection_count_selected_rows (selection);
  /**
   * Loop over selection
   */
  for (; processed_row_count < selected_row_count; processed_row_count++)
  {
    selected_row_path = g_list_nth_data (selected_rows, processed_row_count);
    gtk_tree_model_get_iter (model, &iter, selected_row_path);
    gtk_tree_model_get (model, &iter, COL_DOWNLOAD, &sheet_url, -1);
    gtk_tree_model_get (model, &iter, COL_NAME, &sheet_name, -1);
    percent =
      ((gdouble) processed_row_count) / ((gdouble) selected_row_count);
    gtk_progress_bar_set_fraction (progress, percent);
    message = g_strdup_printf ("Downloading %s", sheet_name);

    gtk_progress_bar_set_fraction (progress, percent);
    gtk_label_set_text (label, message);
    g_free (message);
    if(fetch_sheet (sheet_url, GTK_WINDOW (assistant)))
    {
      extract (progress,
	       label, (processed_row_count + 1) / selected_row_count,
	       GTK_WINDOW (assistant));
      message = g_object_get_data (G_OBJECT (assistant), "sheets_installed");
      if (0 == strncmp (message, "", 1))
      {
	g_object_set_data (G_OBJECT (assistant), "sheets_installed",
			   g_strdup (sheet_name));
      }
      else
      {
	g_object_set_data (G_OBJECT (assistant), "sheets_installed",
			   g_strdup_printf ("%s,%s", message, sheet_name));
      }
    }
    else
    {
      /**
       * \TODO: error handling
       */
    }
  }
  g_list_foreach (selected_rows, gtk_tree_path_free, NULL);
  g_list_free (selected_rows);
  gtk_assistant_set_page_complete (assistant, page, TRUE);
  summary_label = g_object_get_data (G_OBJECT (assistant), "summary_label");
  gtk_label_set_text (summary_label,
		      g_strdup_printf
		      ("The following sheets were installed:\n %s",
		       g_object_get_data (G_OBJECT (assistant),
					  "sheets_installed")));
  gtk_assistant_set_current_page (assistant, 3);
}

/** 
 * Close and destroy the assistant. 
 */
static void
assistant_close (GtkAssistant * assistant, gpointer data)
{
  gtk_widget_destroy (GTK_WIDGET (assistant));
}

static gchar *
get_dia_dir ()
{
  gchar *diadir;
  diadir = g_build_path (G_DIR_SEPARATOR_S, g_get_home_dir (), ".dia", NULL);
  g_mkdir_with_parents(diadir, 755);
  return diadir;
}

/**
 * \TODO: Error handling.
 */
static void 
extract_gsf_infile(GsfInfile *input, gchar *path, GtkWindow *assistant)
{
  gint num_entries = 0;
  gint i = 0;
  GsfInfile *inputchild = NULL;
  GsfOutfile *outputchild = NULL;
  gchar *childpath = NULL;

  num_entries = gsf_infile_num_children(input, NULL);
  for(i = 0; i < num_entries; i++)
  {
        inputchild = gsf_infile_child_by_index(input, i);
        childpath = g_build_filename(path, gsf_input_name(inputchild), NULL);
        if(0 < gsf_infile_num_children(inputchild, NULL))
        {
          if(g_file_test(childpath, G_FILE_TEST_IS_DIR)||
            (0 == g_mkdir_with_parents(childpath, 755)))
          {
               extract_gsf_infile(inputchild, childpath, assistant);
          }
        }
        else
        {
            if(!g_file_test(childpath, G_FILE_TEST_EXISTS))
            {
             outputchild = gsf_output_stdio_new(childpath, NULL);
             gsf_input_copy(inputchild, outputchild);
             g_object_unref(outputchild);
            }
        }
        g_object_unref(inputchild);
        g_free(childpath);
  } 
}

static void
extract (
	 GtkProgressBar * progress, GtkLabel * label, gdouble target_fraction,
	 GtkWindow * assistant)
{
  struct archive *a;
  struct archive *ext;
  struct archive_entry *entry;
  int r;
  int num_entries = 0;
  int i = 0;
  gdouble base_fraction;
  gdouble fraction;
  gchar *message = NULL;
  GtkMessageDialog *dialog = NULL;
  GsfInput *gsfinput = NULL;
  GsfInfile *gsfinfile = NULL;
  GError *error = NULL;
  GsfInput *child = NULL;

  base_fraction = gtk_progress_bar_get_fraction (progress);
  
  gsfinput = gsf_input_memory_new(zip_in_memory, zip_size, TRUE);
  gsfinfile = gsf_infile_zip_new(gsfinput, NULL);
  extract_gsf_infile(gsfinfile, get_dia_dir(), assistant);
  g_object_unref(G_OBJECT(gsfinfile));
  g_free(zip_in_memory);
  zip_in_memory = NULL;
  zip_size = 0;
}

/**
 * Cleanup data
 */
static void
cleanup ()
{
  /**
   * \TODO: remove sheets.xml, if necessary.
   */
}

int
main (int argc, char *argv[])
{
  gchar *sheetpath;

  GtkWidget *assistant, *entry, *label, *button, *progress, *hbox, *vbox,
    *view;
  GtkWidget *scrolledwindow;
  GtkLabel *summary_label = NULL;
  GtkMessageDialog *dialog = NULL;
  guint i;
  PageInfo page[5] = {
    {NULL, -1, "Introduction", GTK_ASSISTANT_PAGE_INTRO, TRUE},
    {NULL, -1, "Select the Shapes", GTK_ASSISTANT_PAGE_CONTENT, FALSE},
    {NULL, -1, "Installation", GTK_ASSISTANT_PAGE_PROGRESS, FALSE},
    {NULL, -1, "Summary", GTK_ASSISTANT_PAGE_SUMMARY, TRUE}
  };

  curl_global_init (CURL_GLOBAL_ALL);
  curl = curl_easy_init ();
  gtk_init (&argc, &argv);
  gsf_init();

  /**
   * Download the repository data.
   */
  sheetpath =
    g_build_path (G_DIR_SEPARATOR_S, get_dia_dir (), "sheets.xml", NULL);

  if (-1 ==
      xmlNanoHTTPFetch ("http://dia-installer.de/sheets.xml", sheetpath,
			NULL))
  {
    dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (NULL,
							 GTK_DIALOG_DESTROY_WITH_PARENT,
							 GTK_MESSAGE_ERROR,
							 GTK_BUTTONS_CLOSE,
							 "Failed to download http://dia-installer.de/sheets.xml.\nTried to save data in '%s'.",
							 sheetpath));
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (GTK_WIDGET (dialog));
    curl_easy_cleanup (curl);
    exit (1);
  }

  /* Create a new assistant widget with no pages. */
  assistant = gtk_assistant_new ();
  gtk_widget_set_size_request (assistant, 450, 300);
  gtk_window_set_title (GTK_WINDOW (assistant), "Dia Shapes Assistant");
  g_signal_connect (G_OBJECT (assistant), "destroy",
		    G_CALLBACK (gtk_main_quit), NULL);

  page[0].widget =
    gtk_label_new (g_strdup_printf
		   ("Welcome to the Dia Shapes Assistant. This assistant\n"
		    "allows you to install additional shapes for Dia\n"
		    "to the folder '%s'\nThe files will be downloaded from the Dia Shape Repository.",
		    get_dia_dir ()));
  page[1].widget = gtk_hbox_new (FALSE, 5);
  page[2].widget = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
  summary_label = gtk_label_new ("The following shapes were installed:\n");
  page[3].widget = GTK_WIDGET (summary_label);


  /**
   * Create the selection widgets for the second page. 
   */
  entry = gtk_entry_new ();
  view = create_view_and_model (sheetpath, GTK_WINDOW (assistant));
  scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (scrolledwindow, view);
  gtk_box_pack_start (GTK_BOX (page[1].widget), scrolledwindow, TRUE, TRUE,
		      5);

   /**
    * Create the progress bar and cancel button, attach
    * to the GtkAlignment widget for later access.
    */
  button = gtk_button_new_with_label ("Install");
  progress = gtk_progress_bar_new ();
  vbox = gtk_vbox_new (FALSE, 2);
  hbox = gtk_hbox_new (FALSE, 5);
  /**
   * \TODO: Make the progress bar work, re-add
   * gtk_box_pack_start (GTK_BOX (hbox), progress, TRUE, FALSE, 5);
   */
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 5);
  label = gtk_label_new ("");
  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, FALSE, 5);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, FALSE, 5);
  gtk_container_add (GTK_CONTAINER (page[2].widget), vbox);
  g_object_set_data (G_OBJECT (page[2].widget), "label", (gpointer) label);
  g_object_set_data (G_OBJECT (page[2].widget), "pbar", (gpointer) progress);
  g_object_set_data (G_OBJECT (page[2].widget), "view", (gpointer) view);
  g_object_set_data (G_OBJECT (page[2].widget), "assistant",
		     (gpointer) assistant);
  g_object_set_data (G_OBJECT (assistant), "sheets_installed", "");
  g_object_set_data (G_OBJECT (assistant), "summary_label", summary_label);

  /**
   * Add pages to the GtkAssistant dialog. 
   */
  for (i = 0; i < 4; i++)
  {
    page[i].index = gtk_assistant_append_page (GTK_ASSISTANT (assistant),
					       page[i].widget);
    gtk_assistant_set_page_title (GTK_ASSISTANT (assistant),
				  page[i].widget, page[i].title);
    gtk_assistant_set_page_type (GTK_ASSISTANT (assistant),
				 page[i].widget, page[i].type);

    /**
     * Set the introduction and summary pages as complete so they can be
     * incremented or closed. 
     */
    gtk_assistant_set_page_complete (GTK_ASSISTANT (assistant),
				     page[i].widget, page[i].complete);
  }

  /**
   * Update whether pages are complete 
   */
  g_signal_connect (G_OBJECT (view), "cursor-changed",
		    G_CALLBACK (cursor_changed), (gpointer) assistant);
  g_signal_connect (G_OBJECT (assistant), "cancel",
		    G_CALLBACK (assistant_cancel), NULL);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (fetch_sheets), page[2].widget);
  g_signal_connect (G_OBJECT (assistant), "close",
		    G_CALLBACK (assistant_close), NULL);
  g_signal_connect (G_OBJECT (assistant), "prepare",
		    G_CALLBACK (assistant_prepare), NULL);

  gtk_widget_show_all (assistant);
  gtk_main ();
  curl_easy_cleanup (curl);
  return 0;
}

static gboolean
fetch_sheet (const gchar * sheet_url, GtkWindow * assistant)
{
  GtkMessageDialog *dialog = NULL;
  GError *error = NULL;
  gchar *content_type = NULL;
  gchar curl_error[CURL_ERROR_SIZE];

  zip_size = 0;
    if (curl)
    {
      curl_easy_setopt (curl, CURLOPT_URL, sheet_url);
      curl_easy_setopt (curl, CURLOPT_NOPROGRESS, 1);
      curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, write_data);
      curl_easy_setopt (curl, CURLOPT_WRITEDATA, NULL);
      curl_easy_setopt (curl, CURLOPT_HEADER, 0);
      curl_easy_setopt (curl, CURLOPT_ERRORBUFFER, &curl_error);
      curl_easy_setopt (curl, CURLOPT_CRLF, 0);
      if (CURLE_OK != curl_easy_perform (curl))
      {
	dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (assistant,
							     GTK_DIALOG_DESTROY_WITH_PARENT,
							     GTK_MESSAGE_ERROR,
							     GTK_BUTTONS_CLOSE,
							     "Failed to fetch %s: %s\n",
							     sheet_url,
							     curl_error));
	g_printerr ("Failed to fetch %s\n", sheet_url);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (GTK_WIDGET (dialog));
	                   return FALSE;
      }
    }
    else
    {
      dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (assistant,
							   GTK_DIALOG_DESTROY_WITH_PARENT,
							   GTK_MESSAGE_ERROR,
							   GTK_BUTTONS_CLOSE,
							   "Curl library initialization failed."));
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (GTK_WIDGET (dialog));
      return FALSE;
    }
 return TRUE;
}

static size_t
write_data (void *ptr, size_t size, size_t nmemb, void *stream)
{
  zip_in_memory = g_realloc(zip_in_memory, zip_size + (size*nmemb));         
  g_memmove(zip_in_memory + zip_size, ptr, size*nmemb);
  zip_size += (size*nmemb);
  return size * nmemb;
}

#ifdef G_OS_WIN32

#define __argc      (*__p___argc())	/* count of cmd line args */
#define __argv      (*__p___argv())	/* pointer to table of cmd line args */

int _stdcall
WinMain (struct HINSTANCE__ *hInstance,
	 struct HINSTANCE__ *hPrevInstance, char *lpszCmdLine, int nCmdShow)
{
  return main (__argc, __argv);
}

#endif
